#include <v8.h>
using namespace v8;

void OurWeakReferenceCallback(
		Persistent<Value> object
		, void* parameter
	) {
	HandleScope scope;
	Persistent<Function>* callback_ptr = ((Persistent<Function>*)parameter);
	Persistent<Function> callback = *callback_ptr;

	Handle<Value> values[0];
	values[0] = object;
	callback->Call(Context::GetCurrent()->Global(),1,values);
	if(object.IsNearDeath()) {
		object.Dispose();
		callback.Dispose();
		delete callback_ptr;
	}
	scope.Close(Undefined());
}

Handle<Value> Weak(
		const Arguments& args
	) {
	HandleScope scope;
	Handle<Value> obj = args[0];
	if(!obj->IsObject() || obj->IsNull() || obj->IsUndefined()) {
		return scope.Close(ThrowException(Exception::Error(String::New("First argument must be an object"))));
	}
	Handle<Value> onCollect = args[1];
	if(onCollect->IsFunction()) {
		Persistent<Object> persistent_obj;
		persistent_obj = Persistent<Object>::New(Handle<Object>::Cast(obj));
		Persistent<Function>* persistent_cb = new Persistent<Function>();
		*persistent_cb = Persistent<Function>::New(Handle<Function>::Cast(onCollect));

		persistent_obj.MakeWeak(
			persistent_cb
			, OurWeakReferenceCallback
		);
		return scope.Close(obj);
	}
	else {
		return scope.Close(ThrowException(Exception::Error(String::New("Collect callback must be a function"))));
	}
}

extern "C" void init (Handle<Object> target)
{
	HandleScope scope;
	
	Local<FunctionTemplate> weak_template = FunctionTemplate::New(Weak);
	Local<Function> weak = weak_template->GetFunction();
	//Export
	target->Set(String::New("OnCollect"),weak);

	scope.Close(Undefined());
}
